/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * proxy_applet.c
 * Copyright (C) Andrea Fabrizi 2009 <andrea.fabrizi@gmail.com>
 * 
 * proxy_applet.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * proxy_applet.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <gtk/gtk.h>
#include <panel-applet.h>
#include <gconf/gconf-client.h>
#include <math.h>
#include "proxy_applet.h"


static const 
BonoboUIVerb proxy_applet_menu_verbs[] = {
        BONOBO_UI_UNSAFE_VERB ("ProxyAppletAbout", proxy_applet_about_cb),
        BONOBO_UI_VERB_END
};

/*
 * About window
 */
static void
proxy_applet_about_cb(BonoboUIComponent *uic, gpointer data, const gchar *verbname)
{
    const char *authors[] =
    {
        "Andrea Fabrizi <andrea.fabrizi@gmail.com>",
        NULL
    };

    gtk_show_about_dialog (NULL,
                           "version", VERSION,
                           "copyright", "Copyright 2009 - Andrea Fabrizi",
                           "comments", "Proxy Applet is a GNOME applet for changing your network proxy configuration with a single click.",
                           "authors", authors,
                           "documenters", NULL,
                           "translator-credits", NULL,
                           "website", "http://www.andreafabrizi.it/?proxyapplet",
                           "website-label", "Proxy Applet Website",
                           "logo-icon-name", GNOME_NETWORK_PROPERTIES,
                           NULL);

}


/*
 * Return pointer to last component of a pathname
 */
gchar *
basename (gchar *path)
{
    gint len, i;
    len = strlen(path);

    if (len <= 1)
        return NULL;

    for (i=len-1; i>=0; i--) {
        if (path[i] == '/')
            return path + i + 1;
    }

    return path;
}

/*
 * GCompareFunc for g_slist_sort
 */
gint
g_slist_sort_callback (gconstpointer a, gconstpointer b)
{
    if ((a == NULL) || (b == NULL))
        return 0;
        
    return g_ascii_strcasecmp((gchar *)a, (gchar *)b);
}

/*
 * Create %key in %dst_dir with value set to %value
 * Return:
 *  TRUE:  OK
 *  FALSE: KO
 */
gboolean
gconf_create_key (GConfClient *gclient, gchar *dst_dir, gchar *key, GConfValue *value)
{
    gchar *new_key;
    gboolean ret;
    
    new_key = g_strconcat(dst_dir, "/", key, NULL); 
    
    if (value->type == GCONF_VALUE_STRING) {
#ifdef DEBUG 
        g_debug ("%s=%s\n",new_key, gconf_value_get_string(value));
#endif
        ret = gconf_client_set_string(gclient, new_key, gconf_value_get_string(value), NULL);
        if (!ret)
            return FALSE;
    }
    else
    if (value->type == GCONF_VALUE_INT) {
#ifdef DEBUG 
        g_debug ("%s=%d\n",key, gconf_value_get_int(value));
#endif
        ret = gconf_client_set_int(gclient, new_key, gconf_value_get_int(value), NULL);
        if (!ret)
            return FALSE;
    }
    else
    if (value->type == GCONF_VALUE_BOOL) {
#ifdef DEBUG 
        g_debug ("%s=%d\n",key, gconf_value_get_bool(value));
#endif
        ret = gconf_client_set_bool(gclient, new_key, gconf_value_get_bool(value), NULL);
        if (!ret)
            return FALSE;
    }
    else
    if (value->type == GCONF_VALUE_LIST) {
        GSList *glist,
               *slist = NULL,
               *ptr;
        
        glist = gconf_value_get_list(value);
        ptr = glist;
        while (ptr) {
            GConfValue *gvalue = ptr->data;

            /* To semplify the code, i assume that this is a list of strings */
            gchar *str_data =  (gchar *)gconf_value_get_string(gvalue);
            slist = g_slist_append(slist, str_data);
            ptr = g_slist_next(ptr);
        }
        
        ret = gconf_client_set_list(gclient, new_key,GCONF_VALUE_STRING, slist, NULL);
        if (!ret)
            return FALSE;
        
        g_slist_free(slist);
        g_slist_free(glist);
    }
           
    g_free(new_key);

    return TRUE;
}

/*
 * Copy all elements from %src_dir to %dst_dir
 * The copy is not recursive!
 * Return:
 *  TRUE:  OK
 *  FALSE: KO
 */
gboolean 
gconf_copy_dir (GConfClient *gclient, gchar *src_dir, gchar *dst_dir)
{
    GSList *src_entries,
           *ptr;
    gboolean ret;

#ifdef DEBUG    
    g_debug ("-> copy gconf %s -> %s\n", src_dir, dst_dir);
#endif

    /* Getting all entries from the source dir */
    src_entries = gconf_client_all_entries(gclient, src_dir, NULL);
    if (src_entries == NULL)
        return FALSE;

    ptr = src_entries;
    while (ptr) {
        GConfEntry *entry =  (GConfEntry *)ptr->data;
        GConfValue *value = entry->value;
        /* Creating new key into dst dir */
        ret = gconf_create_key(gclient, dst_dir, basename((gchar *)entry->key), value);
        g_free(entry->key);
        if (!ret)
            return FALSE;
        ptr = g_slist_next(ptr);
    }
    
    g_slist_free(src_entries);
    
    return TRUE;
}

/*
 * Backup current proxy configuration into GCONF_NETWORK_LOCATIONS
 * Return:
 *  TRUE:  OK
 *  FALSE: KO
 */
gboolean 
proxy_applet_backup_current_location(GConfClient *gclient)
{
    gchar *current_location,
          *dst_location_path,
          *esc_current_location;
    gint ret;

    current_location = gconf_client_get_string(gclient, GCONF_CURRENT_LOCATION, NULL);
    if (current_location == NULL)
        return FALSE;
    
    esc_current_location = gconf_escape_key(current_location, -1);
    g_free(current_location);
    
    dst_location_path = g_strconcat(GCONF_NETWORK_LOCATIONS, "/", esc_current_location, "/", GCONF_HTTP_PROXY_DIR, NULL);
    ret = gconf_copy_dir(gclient, GCONF_SYSTEM_DIR"/"GCONF_HTTP_PROXY_DIR, dst_location_path);
    g_free(dst_location_path);
    if (!ret) {
        g_free(esc_current_location);
        return FALSE;
    }
    
    dst_location_path = g_strconcat(GCONF_NETWORK_LOCATIONS, "/", esc_current_location, "/", GCONF_PROXY_DIR, NULL);
    ret = gconf_copy_dir(gclient, GCONF_SYSTEM_DIR"/"GCONF_PROXY_DIR, dst_location_path);
    g_free(dst_location_path);
    if (!ret) {
        g_free(esc_current_location);
        return FALSE;
    }

    g_free(esc_current_location);
    
    return TRUE;
}

/*
 * Setting up new proxy configuration
 * Return:
 *  TRUE:  OK
 *  FALSE: KO
 */
gboolean
proxy_applet_set_new_location(GConfClient *gclient, gchar *location)
{
    gchar *new_location_path,
          *unesc_current_location;
    gint ret;
    
    new_location_path = g_strconcat(GCONF_NETWORK_LOCATIONS, "/", location, "/", GCONF_HTTP_PROXY_DIR, NULL);
    ret = gconf_copy_dir(gclient, new_location_path, GCONF_SYSTEM_DIR"/"GCONF_HTTP_PROXY_DIR);
    g_free(new_location_path);
    if (!ret)
        return FALSE;
        
    new_location_path = g_strconcat(GCONF_NETWORK_LOCATIONS, "/", location, "/", GCONF_PROXY_DIR, NULL);
    ret = gconf_copy_dir(gclient, new_location_path, GCONF_SYSTEM_DIR"/"GCONF_PROXY_DIR);
    g_free(new_location_path);
    if (!ret)
        return FALSE;

    unesc_current_location = gconf_unescape_key(location, -1);
    if (!gconf_client_set_string(gclient, GCONF_CURRENT_LOCATION, unesc_current_location, NULL)) {
        g_free(unesc_current_location);
        return FALSE;
    }
    
    g_free(unesc_current_location);
    
    return TRUE;
}

/*
 * Remove current proxy configuration from GCONF_NETWORK_LOCATIONS
 * Return:
 *  TRUE:  OK
 *  FALSE: KO
 */
gboolean
proxy_applet_unset_new_location(GConfClient *gclient)
{
    gchar *current_location,
          *unset_dir,
          *esc_current_location;
    gboolean ret;
    
    current_location = gconf_client_get_string(gclient, GCONF_CURRENT_LOCATION, NULL);
    esc_current_location = gconf_escape_key(current_location, -1);
    unset_dir = g_strconcat(GCONF_NETWORK_LOCATIONS, "/", esc_current_location, NULL);

#ifdef DEBUG
    g_debug("-> Unset dir: %s\n", unset_dir);
#endif

    ret = gconf_client_recursive_unset (gclient, unset_dir, (GConfUnsetFlags)0, NULL);
    if (ret == FALSE) {
        g_free(unset_dir);
        g_free(current_location);
        g_free(esc_current_location);
        return FALSE;
    }

    g_free(unset_dir);
    g_free(current_location);    
    g_free(esc_current_location);
    
    return TRUE;
}

/*
 * Switch to new proxy location
 */
gboolean
proxy_applet_switch_location (gchar *location, ProxyApplet *myapplet)
{
    GConfClient *gclient;
    gboolean ret;
        
    gclient = gconf_client_get_default();
    
#ifdef DEBUG 
    g_debug ("Switch location -> %s\n", location);
#endif

    ret = proxy_applet_backup_current_location(gclient);
    if (!ret) {
        g_warning("Unable to backup current location.");
        return FALSE;
    }
    
    proxy_applet_set_new_location(gclient,location);
    if (!ret) {
        g_warning("Unable to set new location.");
        return FALSE;
    }
    
    proxy_applet_unset_new_location(gclient);
    if (!ret) {
        g_warning("Unable to unset new location.");
        return FALSE;
    }
            
    gconf_client_suggest_sync (gclient, NULL);
    g_object_unref(gclient);
    
    return TRUE;
}

/*
 * Menu position function
 */
void 
proxy_applet_popup_position_menu (GtkMenu  *menu, int *x, int *y, gboolean *push_in, gpointer gdata)
{
    GtkWidget      *widget;
    GtkRequisition  requisition;
    gint            menu_xpos;
    gint            menu_ypos;

    widget = GTK_WIDGET (gdata);

    gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

    gdk_window_get_origin (widget->window, &menu_xpos, &menu_ypos);

    menu_xpos += widget->allocation.x;
    menu_ypos += widget->allocation.y;

    switch (panel_applet_get_orient (PANEL_APPLET (widget))) {
        case PANEL_APPLET_ORIENT_DOWN:
        case PANEL_APPLET_ORIENT_UP:
            if (menu_ypos > gdk_screen_get_height (gtk_widget_get_screen (widget)) / 2)
                menu_ypos -= requisition.height;
            else
                menu_ypos += widget->allocation.height;
            break;
        case PANEL_APPLET_ORIENT_RIGHT:
        case PANEL_APPLET_ORIENT_LEFT:
            if (menu_xpos > gdk_screen_get_width (gtk_widget_get_screen (widget)) / 2)
                menu_xpos -= requisition.width;
            else
                menu_xpos += widget->allocation.width;
            break;
        default:
                g_assert_not_reached ();
    }

    *x = menu_xpos;
    *y = menu_ypos;
    *push_in = TRUE;
}

/*
 * MenuItem toggle event handler
 */
gboolean
proxy_applet_menu_toggle (GtkMenuItem *menuitem, gpointer data)
{
    gchar *esc_current_location;
    
    if (!GTK_CHECK_MENU_ITEM (menuitem)->active)
        return TRUE;

    esc_current_location = gconf_escape_key((gchar *)gtk_menu_item_get_label(menuitem), -1);
    
    if (!proxy_applet_switch_location(esc_current_location, (ProxyApplet *)data))
        g_warning("Error switching to location");

    g_free(esc_current_location);
       
    return FALSE;
}

/*
 * Settings toggle
 */
void
proxy_applet_activate_settings (GtkMenuItem *menuitem, gpointer data)
{
    ProxyApplet *myapplet;
    myapplet = (ProxyApplet *)data;
    
    gdk_spawn_command_line_on_screen (gtk_widget_get_screen (GTK_WIDGET (myapplet->applet)), 
                                      GNOME_NETWORK_PROPERTIES, 
                                      NULL);
}

/*
 * Menu unload handler
 */
void
proxy_applet_menu_unload (GtkWidget *widget, gpointer data)
{
    ProxyApplet *myapplet = (ProxyApplet *)data;
    GSList *ptr;
    
#ifdef DEBUG
    g_debug("Menu unload!");
#endif

    /* Freeing locations */    
    ptr = myapplet->locations;
    while (ptr) {
        g_free(ptr->data);
        ptr = g_slist_next(ptr);
    }

    g_slist_free(myapplet->locations);
}

/*
 * Building popup menu
 */
GtkWidget *
proxy_applet_create_menu(ProxyApplet *myapplet)
{
    GtkWidget *gradio,
              *menu,
              *separator,
              *settings,
              *settings_icon;
    GSList *group = NULL,
           *locations_path = NULL,
           *locations = NULL,
           *ptr = NULL;
    GConfClient *gclient;
    gchar *current_location;

    gclient = gconf_client_get_default();
    menu = gtk_menu_new();

    /* Getting current location */
    current_location = gconf_client_get_string(gclient, GCONF_CURRENT_LOCATION, NULL);
    if ((current_location == NULL) || (!strlen(current_location)))
        current_location = g_strdup("Default");
    
    /* Getting all others locations */
    locations_path = gconf_client_all_dirs(gclient, GCONF_NETWORK_LOCATIONS, NULL);

    /* Adding basenames to locations list */
    ptr = locations_path;
    while (ptr) {
        gchar *name = gconf_unescape_key(basename((gchar *)ptr->data), -1);
        if (!g_str_equal(current_location,name))
            locations = g_slist_append(locations, name);
        g_free(ptr->data); 
        ptr = g_slist_next(ptr);
    }

    /* adding current location to locations list */
    locations = g_slist_append(locations, current_location);

    /* Sorting */
    locations = g_slist_sort(locations, g_slist_sort_callback);

    /* Building menu entries */
    ptr = locations;
    while (ptr) {
        gchar *name;
        name = (gchar *)ptr->data;
        gradio = gtk_radio_menu_item_new_with_label (group, name);

        if (g_str_equal(name, current_location)) {
            gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gradio), TRUE);
        }

        myapplet->menu = menu;
        group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(gradio));
        g_signal_connect(G_OBJECT(gradio), "toggled", G_CALLBACK(proxy_applet_menu_toggle), myapplet);
        gtk_widget_show(gradio);
        gtk_menu_shell_append((GtkMenuShell *)menu, gradio);
        ptr=g_slist_next(ptr);
    }

    /* Separator */
    separator = gtk_separator_menu_item_new();
    gtk_widget_show(separator);
    gtk_menu_shell_append((GtkMenuShell *)menu, separator);

    /* Settings */
    settings = gtk_image_menu_item_new_with_label("Settings");
    settings_icon = gtk_image_new_from_stock(GTK_STOCK_PREFERENCES, GTK_ICON_SIZE_MENU);
    gtk_image_menu_item_set_image((GtkImageMenuItem *)settings, settings_icon);
    gtk_widget_show(settings);
    gtk_menu_shell_append((GtkMenuShell *)menu, settings);
    g_signal_connect(G_OBJECT(settings), "activate", G_CALLBACK(proxy_applet_activate_settings), myapplet);

    myapplet->locations = locations;
    g_signal_connect(G_OBJECT(menu), "hide", G_CALLBACK(proxy_applet_menu_unload), myapplet);

    g_slist_free(locations_path);
    g_object_unref(gclient);
    
    return menu;
}

/*
 * Applet click event handler
 */
gint
proxy_applet_mouse_event(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    if ((event->type == GDK_BUTTON_PRESS) && (event->button == 1)) {
        GtkWidget *menu;
        ProxyApplet *myapplet = (ProxyApplet *)data;
        menu = proxy_applet_create_menu(myapplet);
        gtk_menu_popup ((GtkMenu *)menu, NULL, NULL, proxy_applet_popup_position_menu, myapplet->applet, event->button, event->time);        
    }
    
    return FALSE;
}

/*
 * Monitor for GCONF_CURRENT_LOCATION value change
 */
void
proxy_applet_current_location_monitor (GConfClient *client, const gchar *key, GConfValue  *value, gpointer data)
{
    gchar *current_location;
    
    if (!g_str_equal(key, GCONF_CURRENT_LOCATION))
        return;

    current_location = gconf_unescape_key(gconf_value_get_string(value), -1);
    gtk_label_set_text((GtkLabel *)data, current_location);
    g_free(current_location);
    
#ifdef DEBUG
    g_debug("Current location changed! %s\n", key);
#endif

}

/*
 * main()
 */
static gboolean
applet_factory (PanelApplet *applet, const gchar *iid, gpointer user_data)
{
    GtkWidget *label,
              *icon_image,
              *box;
    ProxyApplet *myapplet;
    GConfClient *gclient;
    gchar *current_location,
          *menu_xml;
    
    if (strcmp (iid, "OAFIID:proxy_appletApplet") != 0)
        return FALSE;

    g_set_application_name ("Proxy Applet");
    myapplet = g_malloc0((sizeof(ProxyApplet)));
    gclient = gconf_client_get_default();

    /* Applet icon */
    icon_image = gtk_image_new_from_stock(GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU);
    gtk_widget_show(icon_image);

    /* Applet current location label */
    current_location = gconf_client_get_string(gclient, GCONF_CURRENT_LOCATION, NULL);
    if ((current_location == NULL) || (!strlen(current_location)))
        current_location = "Default";
    label = gtk_label_new (current_location);

    /* Gconf monitor for current location */
    gconf_client_add_dir(gclient, GCONF_NETWORK_LOCATIONS, GCONF_CLIENT_PRELOAD_ONELEVEL,NULL);
    g_signal_connect (G_OBJECT (gclient), "value_changed", G_CALLBACK (proxy_applet_current_location_monitor), label);
    
    box = gtk_hbox_new(FALSE, 3);
    gtk_box_pack_start(GTK_BOX(box), icon_image, FALSE, FALSE, 0);    
    gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);
    gtk_container_add (GTK_CONTAINER(applet), GTK_WIDGET(box));
    
    myapplet->applet = applet;
    myapplet->label_title = label;
    g_signal_connect(G_OBJECT(applet), "button_press_event", G_CALLBACK(proxy_applet_mouse_event), myapplet);
    gtk_widget_show_all (GTK_WIDGET (applet));

    /* Setup the menu */
    menu_xml = g_strdup_printf(proxy_applet_menu_xml,"About");
    panel_applet_setup_menu(applet, menu_xml, proxy_applet_menu_verbs, (gpointer)applet);
    g_free(menu_xml);

    panel_applet_set_flags(applet, PANEL_APPLET_EXPAND_MINOR);
    panel_applet_set_background_widget (PANEL_APPLET (applet), GTK_WIDGET (applet));
    
#ifdef DEBUG
    g_debug ("Applet started...\n");
#endif

    g_object_unref(gclient);
        
    return TRUE;
}

PANEL_APPLET_BONOBO_FACTORY ("OAFIID:proxy_appletApplet_Factory",
                             PANEL_TYPE_APPLET,
                             "Proxy Applet",
                             "0",
                             applet_factory,
                             NULL);

